use std::time::Duration;

use crate::prelude::*;

use tokio::{
    io::{AsyncWriteExt, BufWriter},
    net::TcpStream,
    sync::mpsc::{Receiver, Sender},
};

pub struct ClientSocketActor {
    sender: Sender<ClientSocketActorMessage>,
    fiber: JoinHandle<Result<ClientSocketActorTermination>>,
}

#[derive(Debug, Clone)]
enum ClientSocketActorMessage {
    Terminate,
}

#[derive(Debug, Clone)]
enum ClientSocketActorTermination {
    ServerClosed,
}

impl ClientSocketActor {
    pub async fn terminate(self) -> Result<()> {
        self.sender
            .send(ClientSocketActorMessage::Terminate)
            .await?;
        self.fiber.await??;
        Ok(())
    }

    pub fn launch(
        writer: BufWriter<TcpStream>,
        updater: Sender<u32>,
        id: u32,
    ) -> Result<ClientSocketActor> {
        let (sender, chan) = tokio::sync::mpsc::channel(32);
        let fiber: JoinHandle<Result<ClientSocketActorTermination>> =
            tokio::spawn(ClientSocketActor::handle_connect(writer, chan, updater, id));
        Ok(ClientSocketActor { sender, fiber })
    }

    async fn handle_connect(
        mut writer: BufWriter<TcpStream>,
        mut messagebox: Receiver<ClientSocketActorMessage>,
        updater: Sender<u32>,
        id: u32,
    ) -> Result<ClientSocketActorTermination> {
        async fn handle(
            messagebox: &mut Receiver<ClientSocketActorMessage>,
            writer: &mut BufWriter<TcpStream>,
        ) -> Result<Option<ClientSocketActorTermination>> {
            writer.write_all("privet\n".as_bytes()).await?;
            writer.flush().await?;
            tracing::info!("Wrote");
            tokio::select! {
                _ = messagebox.recv() => return Ok(Some(ClientSocketActorTermination::ServerClosed)),
                _ = tokio::time::sleep(Duration::from_millis(200)) => (),
            }
            Ok(None)
        }

        loop {
            match handle(&mut messagebox, &mut writer).await {
                Ok(None) => {}
                Ok(Some(termination)) => return Ok(termination),
                Err(err) => {
                    tracing::error!("{}", err);
                    updater.send(id).await?;
                    return Err(err);
                }
            }
        }
    }
}
